import numpy as np

import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import matplotlib

# example 1 involving the decision problem of selecting the best wind farm location based on:
"""
Ziemba, P., Wątróbski, J., Zioło, M., & Karczmarczyk, A. (2017). 
Using the PROSA method in offshore wind farm location problems. Energies, 10(11), 1755.
DOI: https://doi.org/10.3390/en10111755
"""

# example 2 involving the decision problem of selecting the best wind farm location based on:
"""
Kizielewicz, B., Wątróbski, J., & Sałabun, W. (2020). Identification of relevant criteria set 
in the MCDA process—Wind farm location case study. Energies, 13(24), 6548.
"""

from pyrepo_mcda.mcda_methods import CRADIS, AHP, MARCOS, SAW, ARAS, COPRAS, PROMETHEE_II, PROSA_C
from pyrepo_mcda.mcda_methods import TOPSIS, VIKOR, MABAC, EDAS, SPOTIS, WASPAS

from pyrepo_mcda import normalizations as norms
from pyrepo_mcda.additions import rank_preferences
from pyrepo_mcda import correlations as corrs
from pyrepo_mcda import weighting_methods as mcda_weights




# bar (column) chart
def plot_barplot(df_plot, legend_title, num):
    """
    Visualization method to display column chart of alternatives rankings obtained with 
    different methods.

    Parameters
    ----------
        df_plot : DataFrame
            DataFrame containing rankings of alternatives obtained with different methods.
            The particular rankings are included in subsequent columns of DataFrame.
        title : str
            Title of the legend (Name of group of explored methods, for example MCDA methods or Distance metrics).
    
    Examples
    ----------
    >>> plot_barplot(df_plot, legend_title='MCDA methods')
    """
    step = 1
    list_rank = np.arange(1, len(df_plot) + 1, step)

    ax = df_plot.plot(kind='bar', width = 0.8, stacked=False, edgecolor = 'black', figsize = (9,4))
    ax.set_xlabel('Alternatives', fontsize = 12)
    ax.set_ylabel('Rank', fontsize = 12)
    ax.set_yticks(list_rank)

    ax.set_xticklabels(df_plot.index, rotation = 'horizontal')
    ax.tick_params(axis = 'both', labelsize = 12)
    y_ticks = ax.yaxis.get_major_ticks()
    ax.set_ylim(0, len(df_plot) + 1)

    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc='lower left',
    ncol=4, mode="expand", borderaxespad=0., edgecolor = 'black', title = legend_title, fontsize = 12)

    ax.grid(True, linestyle = ':')
    ax.set_axisbelow(True)
    plt.tight_layout()
    legend_title = legend_title.replace("$", "")
    plt.savefig('./results/' + 'bar_chart_' + legend_title + str(num) + '.eps')
    plt.savefig('./results/' + 'bar_chart_' + legend_title + str(num) + '.png')
    plt.show()


# heat maps with correlations
def draw_heatmap(df_new_heatmap, title, num):
    """
    Visualization method to display heatmap with correlations of compared rankings generated using different methods
    
    Parameters
    ----------
        data : DataFrame
            DataFrame with correlation values between compared rankings
        title : str
            title of chart containing name of used correlation coefficient
    
    Examples
    ---------
    >>> draw_heatmap(df_new_heatmap, title)
    """
    plt.figure(figsize = (8, 5))
    sns.set(font_scale = 1.2)
    heatmap = sns.heatmap(df_new_heatmap, annot=True, fmt=".3f", cmap="RdYlGn",
                          linewidth=0.5, linecolor='w')
    plt.yticks(va="center")
    plt.xlabel('MCDA methods')
    plt.title('Correlation: ' + title)
    plt.tight_layout()
    title = title.replace("$", "")
    plt.savefig('./results/' + 'correlations_' + title + str(num) + '.eps')
    plt.savefig('./results/' + 'correlations_' + title + str(num) + '.png')
    plt.show()


# Create dictionary class
class Create_dictionary(dict):
  
    # __init__ function
    def __init__(self):
        self = dict()
          
    # Function to add key:value
    def add(self, key, value):
        self[key] = value


def main():

    # Example 1
    # Decision matrix
    matrix = np.array([[16347, 14219, 8160, 8160],
    [9, 8.5, 9, 8.5],
    [73.8, 55, 64.8, 62.5],
    [36.7, 36, 28.5, 29.5],
    [1.5, 2, 2, 1.5],
    [3730, 3240, 1860, 1860],
    [2, 1, 2, 1],
    [1, 1, 2, 3],
    [38.8, 33.1, 45.8, 27.3],
    [4, 2, 4, 3],
    [1720524, 1496512, 858830, 858830],
    [40012, 34803, 19973, 19973]])

    # Transpose the decision matrix to place alternatives in rows and criteria in columns
    matrix = matrix.T

    # Provide weights
    weights = np.array([20, 5, 5, 1.67, 1.67, 11.67, 11.67, 5, 5, 16.67, 8.33, 8.33])
    # Normalize weight values using sum normalization
    # All weights must sum to 1
    weights = weights / np.sum(weights)

    # provide criteria types. 1 represents profit criteria and -1 denotes cost criteria.
    # Other values are not allowed.
    types = np.array([-1, -1, -1, -1, -1, 1, -1, -1, 1, -1, 1, 1])

    # Rows names
    alt_names = [r'$A_{' + str(el) + '}$' for el in range(1, matrix.shape[0] + 1)]
    # Columns names
    col_names = [r'$C_{' + str(el) + '}$' for el in range(1, matrix.shape[1] + 1)]
    # save the DataFrame with dataset to csv file
    df_ex1 = pd.DataFrame(matrix, index = alt_names, columns = col_names)
    df_ex1 = df_ex1.rename_axis('Ai')
    df_ex1.to_csv('./results/dataset1.csv')

    # Create the DataFrame for rankings
    rank_results = pd.DataFrame(index=alt_names)

    
    # PROMETHEE II
    p = np.array([7280, 4, 13.4, 7.4, 3, 1662, 3, 3, 13.8, 3, 766240, 17820])
    promethee_II = PROMETHEE_II()
    preference_functions = [promethee_II._vshape_function for pf in range(len(weights))]

    pref = promethee_II(matrix, weights, types, preference_functions, p = p)
    rank = rank_preferences(pref, reverse=True)
    rank_results['PROMETHEE II'] = rank

    # PROSA-C
    s = np.repeat(0.3, len(weights))
    prosa_c = PROSA_C()
    pref = prosa_c(matrix, weights, types, preference_functions, p = p, s = s)
    rank = rank_preferences(pref, reverse=True)
    rank_results['PROSA C'] = rank

    # ARAS sum norm
    aras = ARAS(normalization_method=norms.linear_normalization)
    pref = aras(matrix, weights, types)
    rank = rank_preferences(pref, reverse=True)
    rank_results['ARAS'] = rank

    # COPRAS sum norm
    copras = COPRAS()
    pref = copras(matrix, weights, types)
    rank = rank_preferences(pref, reverse=True)
    rank_results['COPRAS'] = rank

    # CRADIS linear norm
    cradis = CRADIS()
    pref = cradis(matrix, weights, types)
    rank = rank_preferences(pref, reverse=True)
    rank_results['CRADIS'] = rank

    # MARCOS
    marcos = MARCOS()
    pref = marcos(matrix, weights, types)
    rank = rank_preferences(pref, reverse=True)
    rank_results['MARCOS'] = rank

    # SAW linear norm
    saw = SAW()
    pref = saw(matrix, weights, types)
    rank = rank_preferences(pref, reverse=True)
    rank_results['SAW'] = rank

    # VIKOR
    vikor = VIKOR()
    pref = vikor(matrix, weights, types)
    rank = rank_preferences(pref, reverse=False)
    rank_results['VIKOR'] = rank
    rank_results = rank_results.rename_axis('Ai')
    rank_results.to_csv('./results/results_1.csv')


    print(rank_results)
    plot_barplot(rank_results, legend_title='MCDA methods', num = 1)



    method_types = list(rank_results.columns)
    dict_new_heatmap_rw = Create_dictionary()
    for el in method_types:
        dict_new_heatmap_rw.add(el, [])


    # heatmaps for correlations coefficients
    for i, j in [(i, j) for i in method_types[::-1] for j in method_types]:
        dict_new_heatmap_rw[j].append(corrs.weighted_spearman(rank_results[i], rank_results[j]))

    df_new_heatmap_rw = pd.DataFrame(dict_new_heatmap_rw, index = method_types[::-1])
    df_new_heatmap_rw.columns = method_types

    # correlation matrix with rw coefficient
    draw_heatmap(df_new_heatmap_rw, r'$r_w$', num = 1)

    # Example 2

    matrix = np.array([[106.78, 6.75, 2.00, 220, 6.00, 1, 52.00, 455.50, 8.90, 36.80],
    [86.37, 7.12, 3.00, 400, 10.00, 0, 20.00, 336.50, 7.20, 29.80],
    [104.85, 6.95, 60.00, 220, 7.00, 1, 60.00, 416.00, 8.70, 36.20],
    [46.60, 6.04, 1.00, 220, 3.00, 0, 50.00, 277.00, 3.90, 16.00],
    [69.18, 7.05, 33.16, 220, 8.00, 0, 35.49, 364.79, 5.39, 33.71],
    [66.48, 6.06, 26.32, 220, 6.53, 0, 34.82, 304.02, 4.67, 27.07],
    [74.48, 6.61, 48.25, 400, 4.76, 1, 44.19, 349.45, 4.93, 28.89],
    [73.67, 6.06, 19.54, 400, 3.19, 0, 46.41, 354.65, 8.01, 21.09],
    [100.58, 6.37, 39.27, 220, 8.43, 1, 22.07, 449.42, 7.89, 17.62],
    [94.81, 6.13, 50.58, 220, 4.18, 1, 21.14, 450.88, 5.12, 17.30],
    [48.93, 7.12, 21.48, 220, 5.47, 1, 55.72, 454.71, 8.39, 19.16],
    [74.75, 6.58, 7.08, 400, 9.90, 1, 26.01, 455.17, 4.78, 18.44]])

    types = np.array([1, 1, -1, 1, -1, -1, 1, -1, -1, 1])

    weights = mcda_weights.entropy_weighting(matrix)

    alt_names = [r'$A_{' + str(el) + '}$' for el in range(1, matrix.shape[0] + 1)]
    col_names = [r'$C_{' + str(el) + '}$' for el in range(1, matrix.shape[1] + 1)]
    df_ex2 = pd.DataFrame(matrix, index = alt_names, columns = col_names)
    df_ex2 = df_ex2.rename_axis('Ai')
    df_ex2.to_csv('./results/dataset2.csv')

    rank_results = pd.DataFrame(index=alt_names)

    promethee_II = PROMETHEE_II()
    preference_functions = [prosa_c._vshape_function for pf in range(len(weights))]

    pref = promethee_II(matrix, weights, types, preference_functions)
    rank = rank_preferences(pref, reverse=True)
    rank_results['PROMETHEE II'] = rank

    prosa_c = PROSA_C()

    u = np.sqrt(np.sum(np.square(np.mean(matrix, axis = 0) - matrix), axis = 0) / matrix.shape[0])
    p = 2 * u
    q = 0.5 * u
    s = np.repeat(0.3, len(weights))

    pref = prosa_c(matrix, weights, types, preference_functions, p = p, q = q, s = s)
    rank = rank_preferences(pref, reverse=True)
    rank_results['PROSA C'] = rank

    topsis = TOPSIS()
    pref = topsis(matrix, weights, types)
    rank = rank_preferences(pref, reverse=True)
    rank_results['TOPSIS'] = rank

    vikor = VIKOR()
    pref = vikor(matrix, weights, types)
    rank = rank_preferences(pref, reverse=False)
    rank_results['VIKOR'] = rank

    # SPOTIS preferences must be sorted in ascending order
    bounds_min = np.amin(matrix, axis = 0)
    bounds_max = np.amax(matrix, axis = 0)
    bounds = np.vstack((bounds_min, bounds_max))
    spotis = SPOTIS()
    pref = spotis(matrix, weights, types, bounds)
    rank = rank_preferences(pref, reverse = False)
    rank_results['SPOTIS'] = rank

    # EDAS preferences must be sorted in descending order
    edas = EDAS()
    pref = edas(matrix, weights, types)
    rank = rank_preferences(pref, reverse = True)
    rank_results['EDAS'] = rank

    # MABAC preferences must be sorted in descending order
    mabac = MABAC(normalization_method = norms.minmax_normalization)
    pref = mabac(matrix, weights, types)
    rank = rank_preferences(pref, reverse = True)
    rank_results['MABAC'] = rank

    # CRADIS linear norm
    cradis = CRADIS(normalization_method=norms.minmax_normalization)
    pref = cradis(matrix, weights, types)
    rank = rank_preferences(pref, reverse=True)
    rank_results['CRADIS'] = rank
    rank_results = rank_results.rename_axis('Ai')
    rank_results.to_csv('./results/results_2.csv')

    print(rank_results)
    matplotlib.rc_file_defaults()
    plot_barplot(rank_results, legend_title='MCDA methods', num = 2)


    method_types = list(rank_results.columns)
    dict_new_heatmap_rw = Create_dictionary()
    for el in method_types:
        dict_new_heatmap_rw.add(el, [])


    # heatmaps for correlations coefficients
    for i, j in [(i, j) for i in method_types[::-1] for j in method_types]:
        dict_new_heatmap_rw[j].append(corrs.weighted_spearman(rank_results[i], rank_results[j]))

    df_new_heatmap_rw = pd.DataFrame(dict_new_heatmap_rw, index = method_types[::-1])
    df_new_heatmap_rw.columns = method_types

    # correlation matrix with rw coefficient
    draw_heatmap(df_new_heatmap_rw, r'$r_w$', num = 2)




if __name__ == '__main__':
    main()





import numpy as np

from scipy.sparse.linalg import eigs

from .mcda_method import MCDA_method
from ..normalizations import minmax_normalization



class AHP(MCDA_method):
    def __init__(self, normalization_method = minmax_normalization):
        """
        Create the AHP method object

        Parameters
        -----------
            normalization_method : function
                If you use the AHP method to evaluate a matrix containing the numerical 
                values of the performance alternatives and you have a vector of criteria 
                weights containing their numerical values, then you are not using the 
                `classic_ahp` method but a method called `ahp`.
                You have to choose a method to normalize the given decision matrix.
                The default normalization technique is `minmax_normalization`
        """
        self.normalization_method = normalization_method


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.

        Parameters
        ------------
        matrix : ndarray
                Decision matrix with numerical performance values of alternatives. Decision matrix 
                includes m alternatives in rows and n criteria in columns.
        weights: ndarray
            Vector with criteria weights given in numerical values. The sum of weights 
            must be equal to 1.
        types: ndarray
            Criteria types. Profit criteria are represented by 1 and cost by -1.

        Results
        ---------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ----------
        >>> ahp = AHP()
        >>> pref = ahp(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """
        AHP._verify_input_data(matrix, weights, types)
        return AHP._ahp(self, matrix, weights, types, self.normalization_method)


    def _check_consistency(self, X):
        """
        Consistency Check on the Pairwise Comparison Matrix of the Criteria or alternatives

        Parameters
        -----------
            X : ndarray
                matrix of pairwise comparisons

        Examples
        ----------
        >>> PCcriteria = np.array([[1, 1, 5, 3], [1, 1, 5, 3], [1/5, 1/5, 1, 1/3], [1/3, 1/3, 3, 1]])
        >>> ahp = AHP()
        >>> ahp._check_consistency(PCcriteria)
        """
        n = X.shape[1]
        RI = [0, 0, 0.58, 0.90, 1.12, 1.24, 1.32, 1.41, 1.45, 1.49]
        lambdamax = np.amax(np.linalg.eigvals(X).real)
        CI = (lambdamax - n) / (n - 1)
        CR = CI / RI[n - 1]
        print("Inconsistency index: ", CR)
        if CR > 0.1:
            print("The pairwise comparison matrix is inconsistent")


    def _calculate_eigenvector(self, X):
        """
        Compute the Priority Vector of Criteria (weights) or alternatives using Eigenvector method

        Parameters
        -----------
            X : ndarray
                matrix of pairwise comparisons

        Returns
        ---------
            ndarray
                Eigenvector

        Examples
        ----------
        >>> PCM1 = np.array([[1, 5, 1, 1, 1/3, 3],
        [1/5, 1, 1/3, 1/5, 1/7, 1],
        [1, 3, 1, 1/3, 1/5, 1],
        [1, 5, 3, 1, 1/3, 3],
        [3, 7, 5, 3, 1, 7],
        [1/3, 1, 1, 1/3, 1/7, 1]])
        >>> ahp = AHP()
        >>> S = ahp._calculate_eigenvector(PCM1)
        """
        val, vec = eigs(X, k=1)
        eig_vec = np.real(vec)
        S = eig_vec / np.sum(eig_vec)
        S = S.ravel()
        return S


    def _normalized_column_sum(self, X):
        """
        Compute the Priority Vector of Criteria (weights) or alternatives using The normalized column sum method

        Parameters
        -----------
            X : ndarray
                matrix of pairwise comparisons

        Returns
        ---------
            ndarray
                Vector with weights calculated with The normalized column sum method

        Examples
        ----------
        >>> PCM1 = np.array([[1, 5, 1, 1, 1/3, 3],
        [1/5, 1, 1/3, 1/5, 1/7, 1],
        [1, 3, 1, 1/3, 1/5, 1],
        [1, 5, 3, 1, 1/3, 3],
        [3, 7, 5, 3, 1, 7],
        [1/3, 1, 1, 1/3, 1/7, 1]])
        >>> ahp = AHP()
        >>> S = ahp._normalized_column_sum(PCM1)
        """
        return np.sum(X, axis = 1) / np.sum(X)


    def _geometric_mean(self, X):
        """
        Compute the Priority Vector of Criteria (weights) or alternatives using The geometric mean method

        Parameters
        -----------
            X : ndarray
                matrix of pairwise comparisons

        Returns
        ---------
            ndarray
                Vector with weights calculated with The geometric mean method

        Examples
        ----------
        >>> PCM1 = np.array([[1, 5, 1, 1, 1/3, 3],
        [1/5, 1, 1/3, 1/5, 1/7, 1],
        [1, 3, 1, 1/3, 1/5, 1],
        [1, 5, 3, 1, 1/3, 3],
        [3, 7, 5, 3, 1, 7],
        [1/3, 1, 1, 1/3, 1/7, 1]])
        >>> ahp = AHP()
        >>> S = ahp._geometric_mean(PCM1)
        """
        n = X.shape[1]
        numerator = (np.prod(X, axis = 1))**(1 / n)
        denominator = np.sum(numerator)
        return numerator / denominator

    
    def _classic_ahp(self, alt_matrices, weights, calculate_priority_vector_method = None):
        """
        Calculate the global alternative priorities.
        This is a method for classic AHP where you provide matrices with values of pairwise 
        comparisons of alternatives and weights in the form of a priority vector.

        Parameters
        ------------
            alt_matrices : list
                list with matrices including values of pairwise comparisons of alternatives
            weights : ndarray
                priority vector of criteria (weights)
            calculate_priority_vector_method : function
                Method for calculation of the priority vector. It can be chosen from three
                available methods: _calculate_eigenvector, _normalized_column_sum and
                _geometric_mean
                if the user does not provide calculate_priority_vector_method, it is automatically
                set as the default _calculate_eigenvector

        Returns
        ---------
            ndarray
                vector with the global alternative priorities

        Examples
        -----------
        >>> PCcriteria = np.array([[1, 1, 5, 3], [1, 1, 5, 3], 
        [1/5, 1/5, 1, 1/3], [1/3, 1/3, 3, 1]])
        >>> PCM1 = np.array([[1, 5, 1, 1, 1/3, 3],
        [1/5, 1, 1/3, 1/5, 1/7, 1],
        [1, 3, 1, 1/3, 1/5, 1],
        [1, 5, 3, 1, 1/3, 3],
        [3, 7, 5, 3, 1, 7],
        [1/3, 1, 1, 1/3, 1/7, 1]])
        >>> PCM2 = np.array([[1, 7, 3, 1/3, 1/3, 1/3],
        [1/7, 1, 1/3, 1/7, 1/9, 1/7],
        [1/3, 3, 1, 1/5, 1/5, 1/5],
        [3, 7, 5, 1, 1, 1],
        [3, 9, 5, 1, 1, 1],
        [3, 7, 5, 1, 1, 1]])
        >>> PCM3 = np.array([[1, 1/9, 1/7, 1/9, 1, 1/5],
        [9, 1, 1, 1, 5, 3],
        [7, 1, 1, 1, 5, 1],
        [9, 1, 1, 1, 7, 3],
        [1, 1/5, 1/5, 1/7, 1, 1/3],
        [5, 1/3, 1, 1/3, 3, 1]])
        >>> PCM4 = np.array([[1, 1/5, 1/5, 1/3, 1/7, 1/5],
        [5, 1, 1, 3, 1/3, 1],
        [5, 1, 1, 1, 1/3, 1],
        [3, 1/3, 1, 1, 1/7, 1],
        [7, 3, 3, 7, 1, 5],
        [5, 1, 1, 1, 1/5, 1]])

        >>> ahp = AHP()
        >>> ahp._check_consistency(PCcriteria)
        >>> weights = ahp._calculate_eigenvector(PCcriteria)
        >>> alt_matrices = []
        >>> alt_matrices.append(PCM1)
        >>> alt_matrices.append(PCM2)
        >>> alt_matrices.append(PCM3)
        >>> alt_matrices.append(PCM4)

        >>> calculate_priority_vector_method = ahp._calculate_eigenvector
        >>> pref = ahp._classic_ahp(alt_matrices, weights, calculate_priority_vector_method)
        >>> rank = rank_preferences(pref, reverse = True)
        """

        # eigenvector method is default method to calculate priority vector
        if calculate_priority_vector_method is None:
            calculate_priority_vector_method = self._calculate_eigenvector
        # Check consistency of all pairwise comparison matrices for alternatives 
        for alt in alt_matrices:
            self._check_consistency(alt)

        m = alt_matrices[0].shape[0]
        n = len(weights)

        # Calculate priority vector withe selected method
        S = np.zeros((m, n))
        for el, alt in enumerate(alt_matrices):
            S[:, el] = calculate_priority_vector_method(alt)

        # Calculate the global alternative priorities
        # Calculate the weighted matrix
        Sw = S * weights
        # Aggregate the Local Priorities and Rank the Alternatives
        S_final = np.sum(Sw, axis = 1) / np.sum(Sw)

        return S_final


    @staticmethod
    def _ahp(self, matrix, weights, types, normalization_method):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.

        Parameters
        ------------
        matrix : ndarray
                Decision matrix with numerical performance values of alternatives. The decision matrix 
                includes m alternatives in rows and n criteria in columns.
        weights: ndarray
            Vector with criteria weights given in numerical values. The sum of weights 
            must be equal to 1.
        types: ndarray
            Criteria types. Profit criteria are represented by 1 and cost by -1.

        Results
        ---------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ----------
        >>> ahp = AHP()
        >>> pref = ahp(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """
        nmatrix = normalization_method(matrix, types)
        weighted_matrix = nmatrix * weights
        pref = np.sum(weighted_matrix, axis = 1)
        return pref


import numpy as np


# spearman coefficient rs
def spearman(R, Q):
    """
    Calculate Spearman rank correlation coefficient between two vectors

    Parameters
    -----------
        R : ndarray
            First vector containing values
        Q : ndarray
            Second vector containing values

    Returns
    --------
        float
            Value of correlation coefficient between two vectors

    Examples
    ----------
    >>> rS = spearman(R, Q)
    """

    N = len(R)
    denominator = N*(N**2-1)
    numerator = 6*sum((R-Q)**2)
    rS = 1-(numerator/denominator)
    return rS


# weighted spearman coefficient rw
def weighted_spearman(R, Q):
    """
    Calculate Weighted Spearman rank correlation coefficient between two vectors

    Parameters
    -----------
        R : ndarray
            First vector containing values
        Q : ndarray
            Second vector containing values

    Returns
    --------
        float
            Value of correlation coefficient between two vectors

    Examples
    ---------
    >>> rW = weighted_spearman(R, Q)
    """

    N = len(R)
    denominator = N**4 + N**3 - N**2 - N
    numerator = 6 * sum((R - Q)**2 * ((N - R + 1) + (N - Q + 1)))
    rW = 1 - (numerator / denominator)
    return rW


# pearson coefficient
def pearson_coeff(R, Q):
    """
    Calculate Pearson correlation coefficient between two vectors

    Parameters
    -----------
        R : ndarray
            First vector containing values
        Q : ndarray
            Second vector containing values

    Returns
    --------
        float
            Value of correlation coefficient between two vectors

    Examples
    ----------
    >>> corr = pearson_coeff(R, Q)
    """

    numerator = np.sum((R - np.mean(R)) * (Q - np.mean(Q)))
    denominator = np.sqrt(np.sum((R - np.mean(R))**2) * np.sum((Q - np.mean(Q))**2))
    corr = numerator / denominator
    return corr


# rank similarity coefficient WS
def WS_coeff(R, Q):
    """
    Calculate Rank smilarity coefficient between two vectors

    Parameters
    -----------
        R : ndarray
            First vector containing values
        Q : ndarray
            Second vector containing values

    Returns
    --------
        float
            Value of similarity coefficient between two vectors

    Examples
    ----------
    >>> ws = WS_coeff(R, Q)
    """
    
    N = len(R)
    numerator = 2**(-np.float64(R)) * np.abs(R - Q)
    denominator = np.max((np.abs(R - 1), np.abs(R - N)), axis = 0)
    ws = 1 - np.sum(numerator / denominator)
    return ws


import copy
import time

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib
import seaborn as sns

from pyrepo_mcda.mcda_methods import VIKOR
from pyrepo_mcda.mcda_methods import VIKOR_SMAA
from pyrepo_mcda.additions import rank_preferences
from pyrepo_mcda import correlations as corrs
from pyrepo_mcda import normalizations as norms
from pyrepo_mcda import weighting_methods as mcda_weights


# Functions for visualizations

def plot_barplot(df_plot, x_name, y_name, title):
    """
    Display stacked column chart of weights for criteria for `x_name == Weighting methods`
    and column chart of ranks for alternatives `x_name == Alternatives`
    Parameters
    ----------
        df_plot : dataframe
            dataframe with criteria weights calculated different weighting methods
            or with alternaives rankings for different weighting methods
        x_name : str
            name of x axis, Alternatives or Weighting methods
        y_name : str
            name of y axis, Ranks or Weight values
        title : str
            name of chart title, Weighting methods or Criteria
    Examples
    ----------
    >>> plot_barplot(df_plot, x_name, y_name, title)
    """
    
    list_rank = np.arange(1, len(df_plot) + 1, 1)
    stacked = True
    width = 0.5
    if x_name == 'Alternatives':
        stacked = False
        width = 0.8
    elif x_name == 'Alternative':
        pass
    else:
        df_plot = df_plot.T
    ax = df_plot.plot(kind='bar', width = width, stacked=stacked, edgecolor = 'black', figsize = (9,4))
    ax.set_xlabel(x_name, fontsize = 12)
    ax.set_ylabel(y_name, fontsize = 12)

    if x_name == 'Alternatives':
        ax.set_yticks(list_rank)

    ax.set_xticklabels(df_plot.index, rotation = 'horizontal')
    ax.tick_params(axis = 'both', labelsize = 12)

    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc='lower left',
    ncol=4, mode="expand", borderaxespad=0., edgecolor = 'black', title = title, fontsize = 11)

    ax.grid(True, linestyle = '--')
    ax.set_axisbelow(True)
    plt.tight_layout()
    plt.show()


def draw_heatmap(data, title):
    """
    Display heatmap with correlations of compared rankings generated using different methods
    Parameters
    ----------
        data : dataframe
            dataframe with correlation values between compared rankings
        title : str
            title of chart containing name of used correlation coefficient
    Examples
    ----------
    >>> draw_heatmap(data, title)
    """

    plt.figure(figsize = (6, 4))
    sns.set(font_scale=1.0)
    heatmap = sns.heatmap(data, annot=True, fmt=".2f", cmap="RdYlBu",
                          linewidth=0.5, linecolor='w')
    plt.yticks(va="center")
    plt.xlabel('Weighting methods')
    plt.title('Correlation coefficient: ' + title)
    plt.tight_layout()
    plt.show()

def draw_heatmap_smaa(data, title):
    """
    Display heatmap with correlations of compared rankings generated using different methods
    Parameters
    ----------
        data : dataframe
            dataframe with correlation values between compared rankings
        title : str
            title of chart containing name of used correlation coefficient
    Examples
    ----------
    >>> draw_heatmap(data, title)
    """

    sns.set(font_scale=1.0)
    heatmap = sns.heatmap(data, annot=True, fmt=".2f", cmap="RdYlBu_r",
                        linewidth=0.05, linecolor='w')
    plt.yticks(rotation=0)
    plt.ylabel('Alternatives')
    plt.tick_params(labelbottom=False,labeltop=True)
    
    plt.title(title)
    plt.tight_layout()
    plt.show()


def plot_boxplot(data):
    """
    Display boxplot showing distribution of criteria weights determined with different methods.
    Parameters
    ----------
        data : dataframe
            dataframe with correlation values between compared rankings
    Examples
    ---------
    >>> plot_boxplot(data)
    """
    
    df_melted = pd.melt(data)
    plt.figure(figsize = (7, 4))
    ax = sns.boxplot(x = 'variable', y = 'value', data = df_melted, width = 0.6)
    ax.grid(True, linestyle = '--')
    ax.set_axisbelow(True)
    ax.set_xlabel('Criterion', fontsize = 12)
    ax.set_ylabel('Different weights distribution', fontsize = 12)
    plt.tight_layout()
    plt.show()


# Create dictionary class
class Create_dictionary(dict):
  
    # __init__ function
    def __init__(self):
        self = dict()
          
    # Function to add key:value
    def add(self, key, value):
        self[key] = value



# main
def main():

    matrix = np.array([[155.3, 74.0, 340, 673, 456.0, 111, 115, 106, 244, 39.8, 65440],
        [162.2, 79.5, 247, 639, 283.0, 113, 118, 107, 263, 38.8, 60440],
        [112.5, 68.0, 198, 430, 266.0, 98, 105, 91, 230, 38.1, 56575],
        [90.1, 66.0, 150, 360, 201.2, 120, 131, 109, 259, 34.8, 32495],
        [99.4, 77.0, 150, 310, 201.2, 97, 102, 90, 260, 36.4, 45635],
        [89.5, 40.0, 110, 320, 147.5, 111, 123, 99, 226, 34.8, 28425],
        [124.3, 95.0, 125, 247, 187.7, 78, 78, 77, 222, 40.0, 84595],
        [155.3, 79.2, 160, 300, 214.6, 79, 79, 80, 227, 38.4, 105150],
        [162.2, 100.0, 205, 420, 502.9, 120, 124, 115, 402, 40.3, 96440],
        [96.3, 39.2, 100, 395, 134.1, 120, 132, 108, 258, 34.8, 35245],
        [162.2, 100.0, 205, 420, 502.9, 98, 103, 93, 371, 40.8, 127940],
        [102.5, 38.3, 101, 295, 136.1, 133, 145, 121, 170, 34.8, 34250]])

    types = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1])

    

    # matrix = np.array([[80, 16, 2, 5],
    # [110, 32, 2, 9],
    # [130, 64, 4, 9],
    # [185, 64, 4, 1],
    # [135, 64, 3, 4],
    # [140, 32, 3, 5],
    # [185, 64, 6, 7],
    # [110, 16, 3, 3],
    # [120, 16, 4, 3],
    # [340, 128, 6, 5]])

    # types = np.array([-1, 1, 1, 1])

    # Symbols for alternatives Ai
    list_alt_names = [r'$A_{' + str(i) + '}$' for i in range(1, matrix.shape[0] + 1)]
    # Symbols for columns Cj
    cols = [r'$C_{' + str(j) + '}$' for j in range(1, matrix.shape[1] + 1)]

    
    # Part 1 - study with single weighting method
    
    # Determine criteria weights with chosen weighting method
    weights = mcda_weights.entropy_weighting(matrix)

    # Create the VIKOR method object
    # vikor = VIKOR(normalization_method=norms.minmax_normalization)
    vikor = VIKOR(normalization_method=norms.minmax_normalization)
    
    # Calculate alternatives preference function values with VIKOR method
    pref = vikor(matrix, weights, types)

    # when there is only one (single) preference vector
    rank = rank_preferences(pref, reverse = False)

    print(rank)

    
    # Part 2 - study with several weighting methods
    # Create a list with weighting methods that you want to explore
    weighting_methods_set = [
        mcda_weights.entropy_weighting,
        # mcda_weights.std_weighting,
        mcda_weights.critic_weighting,
        mcda_weights.gini_weighting,
        mcda_weights.merec_weighting,
        mcda_weights.stat_var_weighting,
        # mcda_weights.cilos_weighting,
        mcda_weights.idocriw_weighting,
        mcda_weights.angle_weighting,
        mcda_weights.coeff_var_weighting
    ]
    

    #df_weights = pd.DataFrame(weights.reshape(1, -1), index = ['Weights'], columns = cols)
    # Create dataframes for weights, preference function values and rankings determined using different weighting methods
    df_weights = pd.DataFrame(index = cols)
    df_preferences = pd.DataFrame(index = list_alt_names)
    df_rankings = pd.DataFrame(index = list_alt_names)

    # Create the VIKOR method object
    vikor = VIKOR(normalization_method=norms.minmax_normalization)

    for weight_type in weighting_methods_set:
        
        if weight_type.__name__ in ["cilos_weighting", "idocriw_weighting", "angle_weighting", "merec_weighting"]:
            weights = weight_type(matrix, types)
        else:
            weights = weight_type(matrix)

        df_weights[weight_type.__name__[:-10].upper().replace('_', ' ')] = weights
        pref = vikor(matrix, weights, types)

        rank = rank_preferences(pref, reverse = False)

        df_preferences[weight_type.__name__[:-10].upper().replace('_', ' ')] = pref
        df_rankings[weight_type.__name__[:-10].upper().replace('_', ' ')] = rank
        

    
    # plot criteria weights distribution using box chart
    plot_boxplot(df_weights.T)

    # plot stacked column chart of criteria weights
    plot_barplot(df_weights, 'Weighting methods', 'Weight value', 'Criteria')

    # plot column chart of alternatives rankings
    plot_barplot(df_rankings, 'Alternatives', 'Rank', 'Weighting methods')

    # Plot heatmaps of rankings correlation coefficient
    # Create dataframe with rankings correlation values
    results = copy.deepcopy(df_rankings)
    method_types = list(results.columns)
    dict_new_heatmap_rw = Create_dictionary()

    for el in method_types:
        dict_new_heatmap_rw.add(el, [])

    for i, j in [(i, j) for i in method_types[::-1] for j in method_types]:
        dict_new_heatmap_rw[j].append(corrs.weighted_spearman(results[i], results[j]))
            
    df_new_heatmap_rw = pd.DataFrame(dict_new_heatmap_rw, index = method_types[::-1])
    df_new_heatmap_rw.columns = method_types

    # Plot heatmap with rankings correlation
    draw_heatmap(df_new_heatmap_rw, r'$r_w$')
    

    
    # SMAA method
    cols_ai = [str(el) for el in range(1, matrix.shape[0] + 1)]

    # criteria number
    n = matrix.shape[1]
    # SMAA iterations number (number of weight vectors for SMAA)
    iterations = 10000

    start = time.time()
    # create the VIKOR_SMAA object
    vikor_smaa = VIKOR_SMAA()
    # generate matrix with weight vectors for SMAA
    weight_vectors = vikor_smaa._generate_weights(n, iterations)

    # run the VIKOR_SMAA method
    rank_acceptability_index, central_weight_vector, rank_scores = vikor_smaa(matrix, weight_vectors, types)

    end = time.time() - start
    print('Run time: ', end)
    
    acc_in_df = pd.DataFrame(rank_acceptability_index, index = list_alt_names, columns = cols_ai)
    

    matplotlib.rcdefaults()
    plot_barplot(acc_in_df, 'Alternative', 'Rank acceptability index', 'Rank')

    draw_heatmap_smaa(acc_in_df, 'Rank acceptability indexes')

    central_weights_df = pd.DataFrame(central_weight_vector, index = list_alt_names, columns = cols)
    

    rank_scores_df = pd.DataFrame(rank_scores, index = list_alt_names, columns = ['Rank'])
    
    


if __name__ == '__main__':
    main()